import { ASTKindToNode, GraphQLDirective, GraphQLDirectiveConfig, Kind } from 'graphql';
import { GetGen, GetGen2, NexusWrappedSymbol } from '../core';
import type { MaybeReadonlyArray } from '../typeHelpersInternal';
import type { ArgsRecord } from './args';
import { Maybe } from './_types';
export declare type Directives = MaybeReadonlyArray<NexusDirectiveDef | GraphQLDirective | NexusDirectiveUse>;
export declare const RequestDirectiveLocation: string[];
export declare const SchemaDirectiveLocation: readonly ["SCHEMA", "SCALAR", "OBJECT", "FIELD_DEFINITION", "ARGUMENT_DEFINITION", "INTERFACE", "UNION", "ENUM", "ENUM_VALUE", "INPUT_OBJECT", "INPUT_FIELD_DEFINITION"];
export declare type SchemaDirectiveLocationEnum = typeof SchemaDirectiveLocation[number];
export declare type RequestDirectiveLocationEnum = typeof RequestDirectiveLocation[number];
export interface NexusDirectiveConfig<DirectiveName extends string = string> {
    /** Name of the directive */
    name: DirectiveName;
    /** The description to annotate the GraphQL SDL */
    description?: string;
    /** Valid locations that this directive may be used */
    locations: MaybeReadonlyArray<SchemaDirectiveLocationEnum | RequestDirectiveLocationEnum>;
    /** Whether the directive can be repeated */
    isRepeatable?: Maybe<boolean> | undefined;
    /**
     * [GraphQL.org Docs](https://graphql.github.io/learn/schema/#arguments) | [GraphQL 2018
     * Spec](https://spec.graphql.org/June2018/#sec-Language.Arguments)
     *
     * Define arguments for this directive.
     *
     * All directives in GraphQL can have arguments defined for them. Nexus provides a number of helpers for
     * defining arguments. All builtin GraphQL scalar types have helpers named "{scalarName}Arg" such as
     * "stringArg" and "intArg". You can also use type modifier helpers "[list](https://nxs.li/docs/api/list)"
     * "[nullable](https://nxs.li/docs/api/nullable)" and "[nonNull](https://nxs.li/docs/api/nonNull)". For
     * details about nonNull/nullable refer to the [nullability guide](https://nxs.li/guides/nullability).
     *
     * @example
     *   export const TestValue = directive({
     *     name: 'TestValue',
     *     description: 'Denotes the value used when testing this type',
     *     args: {
     *       type: enumType({
     *         name: 'TestValueType',
     *         members: ['String', 'Int', 'Float', 'JSON'],
     *       }),
     *       value: stringArg(),
     *       listLength: intArg(),
     *     },
     *     locations: ['FIELD_DEFINITION'],
     *   })
     */
    args?: ArgsRecord;
    /** Data that will be added to the directive extensions field on the graphql-js type def instances */
    extensions?: GraphQLDirectiveConfig['extensions'];
}
export interface NexusDirectiveDef<DirectiveName extends string = string> {
    (...args: MaybeArgsFor<DirectiveName>): NexusDirectiveUse<DirectiveName>;
    value: NexusDirectiveConfig;
    [NexusWrappedSymbol]: 'Directive';
}
export declare class NexusDirectiveUse<DirectiveName extends string = string> {
    readonly name: DirectiveName;
    readonly args: MaybeArgsFor<DirectiveName>[0] | undefined;
    readonly config?: NexusDirectiveConfig<DirectiveName> | undefined;
    constructor(name: DirectiveName, args: MaybeArgsFor<DirectiveName>[0] | undefined, config?: NexusDirectiveConfig<DirectiveName> | undefined);
}
/** Defines a directive, which can be used positionally when generating the SDL */
export declare function directive<DirectiveName extends string>(config: NexusDirectiveConfig<DirectiveName>): NexusDirectiveDef<DirectiveName>;
declare type MaybeArgsFor<DirectiveName extends string> = GetGen2<'directiveArgs', DirectiveName> extends object ? [GetGen2<'directiveArgs', DirectiveName>] : [];
export declare function addDirective<DirectiveName extends GetGen<'directives', string>>(directiveName: DirectiveName, ...args: MaybeArgsFor<DirectiveName>): NexusDirectiveUse<DirectiveName>;
declare const DirectiveASTKindMapping: {
    readonly SCALAR: Kind.SCALAR_TYPE_DEFINITION;
    readonly SCHEMA: Kind.SCHEMA_DEFINITION;
    readonly OBJECT: Kind.OBJECT_TYPE_DEFINITION;
    readonly FIELD_DEFINITION: Kind.FIELD_DEFINITION;
    readonly ARGUMENT_DEFINITION: Kind.INPUT_VALUE_DEFINITION;
    readonly INTERFACE: Kind.INTERFACE_TYPE_DEFINITION;
    readonly UNION: Kind.UNION_TYPE_DEFINITION;
    readonly ENUM: Kind.ENUM_TYPE_DEFINITION;
    readonly ENUM_VALUE: Kind.ENUM_VALUE_DEFINITION;
    readonly INPUT_OBJECT: Kind.INPUT_OBJECT_TYPE_DEFINITION;
    readonly INPUT_FIELD_DEFINITION: Kind.INPUT_VALUE_DEFINITION;
};
declare type DirectiveASTKindMapping = typeof DirectiveASTKindMapping;
export declare type DirectiveASTKinds = keyof DirectiveASTKindMapping;
/**
 * Creates the ASTNode with the directives
 *
 * @param
 */
export declare function maybeAddDirectiveUses<Kind extends DirectiveASTKinds, Result extends {
    astNode: ASTKindToNode[DirectiveASTKindMapping[Kind]];
}>(kind: Kind, directiveUses: Directives | undefined, customDirectives: Record<string, GraphQLDirective>): Result | undefined;
export {};
