"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.maybeAddDirectiveUses = exports.addDirective = exports.directive = exports.NexusDirectiveUse = exports.SchemaDirectiveLocation = exports.RequestDirectiveLocation = void 0;
const graphql_1 = require("graphql");
const core_1 = require("../core");
const utils_1 = require("../utils");
const wrapping_1 = require("./wrapping");
const _types_1 = require("./_types");
exports.RequestDirectiveLocation = [
    /** Request Definitions */
    'QUERY',
    'MUTATION',
    'SUBSCRIPTION',
    'FIELD',
    'FRAGMENT_DEFINITION',
    'FRAGMENT_SPREAD',
    'INLINE_FRAGMENT',
    'VARIABLE_DEFINITION',
];
exports.SchemaDirectiveLocation = [
    /** Type System Definitions */
    'SCHEMA',
    'SCALAR',
    'OBJECT',
    'FIELD_DEFINITION',
    'ARGUMENT_DEFINITION',
    'INTERFACE',
    'UNION',
    'ENUM',
    'ENUM_VALUE',
    'INPUT_OBJECT',
    'INPUT_FIELD_DEFINITION',
];
class NexusDirectiveUse {
    constructor(name, args, config) {
        this.name = name;
        this.args = args;
        this.config = config;
    }
}
exports.NexusDirectiveUse = NexusDirectiveUse;
(0, _types_1.withNexusSymbol)(NexusDirectiveUse, _types_1.NexusTypes.DirectiveUse);
/** Defines a directive, which can be used positionally when generating the SDL */
function directive(config) {
    (0, graphql_1.assertValidName)(config.name);
    config = Object.freeze(config);
    function addDirective(...args) {
        return new NexusDirectiveUse(config.name, args[0], config);
    }
    addDirective[core_1.NexusWrappedSymbol] = _types_1.NexusTypes.Directive;
    addDirective.value = config;
    return addDirective;
}
exports.directive = directive;
function addDirective(directiveName, ...args) {
    return new NexusDirectiveUse(directiveName, args[0]);
}
exports.addDirective = addDirective;
const DirectiveASTKindMapping = {
    SCALAR: graphql_1.Kind.SCALAR_TYPE_DEFINITION,
    SCHEMA: graphql_1.Kind.SCHEMA_DEFINITION,
    OBJECT: graphql_1.Kind.OBJECT_TYPE_DEFINITION,
    FIELD_DEFINITION: graphql_1.Kind.FIELD_DEFINITION,
    ARGUMENT_DEFINITION: graphql_1.Kind.INPUT_VALUE_DEFINITION,
    INTERFACE: graphql_1.Kind.INTERFACE_TYPE_DEFINITION,
    UNION: graphql_1.Kind.UNION_TYPE_DEFINITION,
    ENUM: graphql_1.Kind.ENUM_TYPE_DEFINITION,
    ENUM_VALUE: graphql_1.Kind.ENUM_VALUE_DEFINITION,
    INPUT_OBJECT: graphql_1.Kind.INPUT_OBJECT_TYPE_DEFINITION,
    INPUT_FIELD_DEFINITION: graphql_1.Kind.INPUT_VALUE_DEFINITION, // 'InputValueDefinition',
};
/**
 * Creates the ASTNode with the directives
 *
 * @param
 */
function maybeAddDirectiveUses(kind, directiveUses, customDirectives) {
    if (!(directiveUses === null || directiveUses === void 0 ? void 0 : directiveUses.length)) {
        return undefined;
    }
    const seenDirectives = new Set();
    return {
        astNode: {
            kind: DirectiveASTKindMapping[kind],
            directives: Array.from(directiveUses).map((d) => {
                const directiveName = (0, wrapping_1.isNexusDirective)(d) ? d.value.name : d.name;
                const directiveDef = customDirectives[directiveName];
                if (seenDirectives.has(directiveName) && !directiveDef.isRepeatable) {
                    throw new Error(`Cannot use directive ${directiveName} more than once in a row`);
                }
                assertValidDirectiveFor(kind, directiveDef);
                seenDirectives.add(directiveName);
                const args = (0, wrapping_1.isNexusDirectiveUse)(d) ? directiveArgs(d, directiveDef) : undefined;
                return {
                    kind: graphql_1.Kind.DIRECTIVE,
                    name: {
                        kind: graphql_1.Kind.NAME,
                        value: directiveName,
                    },
                    arguments: args,
                };
            }),
        },
    };
}
exports.maybeAddDirectiveUses = maybeAddDirectiveUses;
function assertValidDirectiveFor(kind, directiveDef // any is a hack to make this work w/ v15 & 16
) {
    if (!directiveDef.locations.includes(kind)) {
        throw new Error(`Directive ${directiveDef.name} cannot be applied to ${kind}`);
    }
}
function directiveArgs(directiveUse, directiveDef) {
    var _a;
    return (0, utils_1.mapObj)((_a = directiveUse.args) !== null && _a !== void 0 ? _a : {}, (val, key) => {
        const arg = directiveDef.args.find((a) => a.name === key);
        if (!arg) {
            throw new Error(`Unknown directive arg ${key}, expected one of ${directiveDef.args.map((a) => a.name)}`);
        }
        const value = (0, graphql_1.astFromValue)(val, arg.type);
        if (!value) {
            throw new Error(`Unable to get ast for ${key}`);
        }
        return {
            kind: graphql_1.Kind.ARGUMENT,
            name: {
                kind: graphql_1.Kind.NAME,
                value: key,
            },
            value,
        };
    });
}
//# sourceMappingURL=directive.js.map