"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.makeSchemaInternal = exports.generateSchema = exports.makeSchema = void 0;
const tslib_1 = require("tslib");
const graphql_1 = require("graphql");
const wrapping_1 = require("./definitions/wrapping");
const builder_1 = require("./builder");
const typegenMetadata_1 = require("./typegenMetadata");
const typegenUtils_1 = require("./typegenUtils");
const utils_1 = require("./utils");
/**
 * Defines the GraphQL schema, by combining the GraphQL types defined by the GraphQL Nexus layer or any
 * manually defined GraphQLType objects.
 *
 * Requires at least one type be named "Query", which will be used as the root query type.
 */
function makeSchema(config) {
    const { schema, missingTypes, finalConfig, hasSDLDirectives } = makeSchemaInternal(config);
    const typegenConfig = (0, typegenUtils_1.resolveTypegenConfig)(finalConfig);
    const sdl = typegenConfig.outputs.schema;
    const typegen = typegenConfig.outputs.typegen;
    if (sdl || typegen) {
        // Generating in the next tick allows us to use the schema
        // in the optional thunk for the typegen config
        const typegenPromise = new typegenMetadata_1.TypegenMetadata(typegenConfig).generateArtifacts(schema, hasSDLDirectives);
        if (config.shouldExitAfterGenerateArtifacts) {
            let typegenPath = '(not enabled)';
            if (typegenConfig.outputs.typegen) {
                typegenPath = typegenConfig.outputs.typegen.outputPath;
                if (typegenConfig.outputs.typegen.globalsPath) {
                    typegenPath += ` / ${typegenConfig.outputs.typegen.globalsPath}`;
                }
            }
            typegenPromise
                .then(() => {
                console.log(`Generated Artifacts:
          TypeScript Types  ==> ${typegenPath}
          GraphQL Schema    ==> ${typegenConfig.outputs.schema || '(not enabled)'}`);
                process.exit(0);
            })
                .catch((e) => {
                console.error(e);
                process.exit(1);
            });
        }
        else {
            typegenPromise.catch((e) => {
                console.error(e);
            });
        }
    }
    (0, utils_1.assertNoMissingTypes)(schema, missingTypes);
    (0, utils_1.runAbstractTypeRuntimeChecks)(schema, finalConfig.features);
    return schema;
}
exports.makeSchema = makeSchema;
/** Like makeSchema except that typegen is always run and waited upon. */
function generateSchema(config) {
    return (0, tslib_1.__awaiter)(this, void 0, void 0, function* () {
        const { schema, missingTypes, finalConfig, hasSDLDirectives } = makeSchemaInternal(config);
        const typegenConfig = (0, typegenUtils_1.resolveTypegenConfig)(finalConfig);
        yield new typegenMetadata_1.TypegenMetadata(typegenConfig).generateArtifacts(schema, hasSDLDirectives);
        (0, utils_1.assertNoMissingTypes)(schema, missingTypes);
        (0, utils_1.runAbstractTypeRuntimeChecks)(schema, finalConfig.features);
        return schema;
    });
}
exports.generateSchema = generateSchema;
/**
 * Mainly useful for testing, generates the schema and returns the artifacts that would have been otherwise
 * written to the filesystem.
 */
generateSchema.withArtifacts = (config, typegen = null) => (0, tslib_1.__awaiter)(void 0, void 0, void 0, function* () {
    const { schema, missingTypes, finalConfig, hasSDLDirectives } = makeSchemaInternal(config);
    const typegenConfig = (0, typegenUtils_1.resolveTypegenConfig)(finalConfig);
    const { schemaTypes, tsTypes, globalTypes } = yield new typegenMetadata_1.TypegenMetadata(typegenConfig).generateArtifactContents(schema, typegen, hasSDLDirectives);
    (0, utils_1.assertNoMissingTypes)(schema, missingTypes);
    (0, utils_1.runAbstractTypeRuntimeChecks)(schema, finalConfig.features);
    return { schema, schemaTypes, tsTypes, globalTypes };
});
/** Builds the schema, we may return more than just the schema from this one day. */
function makeSchemaInternal(config) {
    const builder = new builder_1.SchemaBuilder(config);
    builder.addTypes(config.types);
    builder.addTypes(config.directives);
    if (config.schemaRoots) {
        builder.addTypes(config.schemaRoots);
    }
    function getRootType(rootType, defaultType) {
        var _a, _b;
        const rootTypeVal = (_b = (_a = config.schemaRoots) === null || _a === void 0 ? void 0 : _a[rootType]) !== null && _b !== void 0 ? _b : defaultType;
        let returnVal = null;
        if (typeof rootTypeVal === 'string') {
            returnVal = typeMap[rootTypeVal];
        }
        else if (rootTypeVal) {
            if ((0, wrapping_1.isNexusObjectTypeDef)(rootTypeVal)) {
                returnVal = typeMap[rootTypeVal.name];
            }
            else if ((0, graphql_1.isObjectType)(rootTypeVal)) {
                returnVal = typeMap[rootTypeVal.name];
            }
        }
        if (returnVal && !(0, graphql_1.isObjectType)(returnVal)) {
            throw new Error(`Expected ${rootType} to be a objectType, saw ${returnVal.constructor.name}`);
        }
        return returnVal;
    }
    const { finalConfig, typeMap, missingTypes, schemaExtension, onAfterBuildFns, customDirectives, schemaDirectives, hasSDLDirectives, } = builder.getFinalTypeMap();
    const schema = new graphql_1.GraphQLSchema(Object.assign(Object.assign(Object.assign({}, extractGraphQLSchemaOptions(config)), { query: getRootType('query', 'Query'), mutation: getRootType('mutation', 'Mutation'), subscription: getRootType('subscription', 'Subscription'), types: (0, utils_1.objValues)(typeMap), extensions: Object.assign(Object.assign({}, config.extensions), { nexus: schemaExtension }), directives: [...graphql_1.specifiedDirectives, ...Object.values(customDirectives)] }), schemaDirectives));
    onAfterBuildFns.forEach((fn) => fn(schema));
    return { schema, missingTypes, finalConfig, hasSDLDirectives };
}
exports.makeSchemaInternal = makeSchemaInternal;
function extractGraphQLSchemaOptions(config) {
    const { formatTypegen, nonNullDefaults, mergeSchema, outputs, shouldExitAfterGenerateArtifacts, shouldGenerateArtifacts, schemaRoots, sourceTypes, prettierConfig, plugins, customPrintSchemaFn, features, contextType } = config, graphqlConfigOpts = (0, tslib_1.__rest)(config, ["formatTypegen", "nonNullDefaults", "mergeSchema", "outputs", "shouldExitAfterGenerateArtifacts", "shouldGenerateArtifacts", "schemaRoots", "sourceTypes", "prettierConfig", "plugins", "customPrintSchemaFn", "features", "contextType"]);
    return graphqlConfigOpts;
}
//# sourceMappingURL=makeSchema.js.map